#include "stdafx.h"

#include "camera.h"

namespace pco
{
  //_________________________________________
  // constructor + destructor
  Camera::Camera(CameraInterface cam_interface)
  {

    //___________________________________________
    std::memset(&camera_description_, 0, sizeof(PCO_Description)); //erase meaningless information
    camera_description_.wSize = sizeof(PCO_Description);      //size of PCO-structs must always be set
    std::memset(&camera_type_, 0, sizeof(PCO_CameraType));
    camera_type_.wSize = sizeof(PCO_CameraType);

    //___________________________________________
    scanForCamera(cam_interface);

    int err = PCO_NOERROR;
    //______________________________________________________________
    //set and get general camera information
    err = PCO_GetCameraDescription(sdk_, &camera_description_);
    if (err)
      throw CameraException(err);

    err = PCO_GetCameraType(sdk_, &camera_type_);
    if (err)
      throw CameraException(err);

    char sz_camera_name[MAX_PATH];
    err = PCO_GetCameraName(sdk_, sz_camera_name, MAX_PATH);
    if (err)
      throw CameraException(err);
    camera_name_ = std::string(sz_camera_name);

    std::vector<std::pair<DataFormat, int>> format2type;
    if (isColored())
    {
      format2type = std::vector<std::pair<DataFormat, int>>{
        {DataFormat::Mono8, PCO_BW_CONVERT},
        {DataFormat::BGR8, PCO_COLOR_CONVERT},
        {DataFormat::BGR16, PCO_COLOR16_CONVERT},
        {DataFormat::BGRA8, 0}
      };

      conv_ctrl_map_[DataFormat::Mono8] = ConvertControlMono();
      conv_ctrl_map_[DataFormat::BGR8] = ConvertControlColor();
      conv_ctrl_map_[DataFormat::BGR16] = ConvertControlColor();
      conv_ctrl_map_[DataFormat::BGRA8] = ConvertControlColor();

      checkMinMaxScale(DataFormat::BGR8);
      checkMinMaxScale(DataFormat::BGRA8);

      do_auto_minmax_map_[DataFormat::Mono8] = true;
      do_auto_minmax_map_[DataFormat::BGR8] = true;
      do_auto_minmax_map_[DataFormat::BGR16] = true;
      do_auto_minmax_map_[DataFormat::BGRA8] = true;
    }
    else
    {
      format2type = std::vector<std::pair<DataFormat, int>>{
        {DataFormat::Mono8, PCO_BW_CONVERT},
        {DataFormat::BGR8, PCO_PSEUDO_CONVERT},
        {DataFormat::BGRA8, 0}
      };

      conv_ctrl_map_[DataFormat::Mono8] = ConvertControlMono();
      conv_ctrl_map_[DataFormat::BGR8] = ConvertControlPseudoColor();
      conv_ctrl_map_[DataFormat::BGRA8] = ConvertControlPseudoColor();

      do_auto_minmax_map_[DataFormat::Mono8] = true;
      do_auto_minmax_map_[DataFormat::BGR8] = true;
      do_auto_minmax_map_[DataFormat::BGRA8] = true;
    }
    PCO_SensorInfo sensor_info = getSensorInfo();
    for (auto& [data_format, type] : format2type)
    {
      if (data_format == DataFormat::BGRA8)
      {
        // same convert handle
        conv_map_[DataFormat::BGRA8] = conv_map_.at(DataFormat::BGR8);
        updateDisplaySettings(data_format);
        continue;
      }
      HANDLE conv = nullptr;
      err = PCO_ConvertCreate(&conv, &sensor_info, type);
      if (err)
        throw CameraException(err);

      conv_map_[data_format] = conv;

      updateDisplaySettings(data_format);

    }


    defaultConfiguration();

  }

  Camera::~Camera()
  {
    if (isRecording())
      stop();
    close();
  }

  //_________________________________________
  // camera properties

  std::string Camera::getName() const
  {
    return camera_name_;
  }

  RawFormat Camera::getRawFormat() const
  {
    if (camera_description_.wDynResDESC > 8)
      return RawFormat::Word;
    else
      return RawFormat::Byte;
  }

  DWORD Camera::getSerial() const
  {
    return camera_type_.dwSerialNumber;
  }

  CameraInterface Camera::getInterface() const
  {
    return interface_;
  }

  bool Camera::isRecording() noexcept
  {
    if (rec_ == nullptr)
      return false;

    int err = PCO_NOERROR;
    bool is_running = false;
    err = PCO_RecorderGetStatus(rec_, sdk_, &is_running, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr);
    if (err)
      return false;

    return is_running;
  }

  bool Camera::isColored() const
  {
    return camera_description_.wSensorTypeDESC & 0x0001;
  }

  //_________________________________________
  // control camera

  Description Camera::getDescription() const
  {
    Description desc;
    desc.serial = camera_type_.dwSerialNumber;
    desc.type = camera_type_.wCamType;
    desc.sub_type = camera_type_.wCamSubType;
    desc.interface_type = interface_;

    int err = PCO_NOERROR;
    if (sdk_ != nullptr)
    {
      PCO_Description pco_desc{ sizeof(PCO_Description) };
      err = PCO_GetCameraDescription(sdk_, &pco_desc);
      if (err)
        throw CameraException(err);

      PCO_Description3 pco_desc_3{ sizeof(PCO_Description3) };
      if (pco_desc.dwGeneralCapsDESC1 & GENERALCAPS1_ENHANCED_DESCRIPTOR_3)
      {
        err = PCO_GetCameraDescriptionEx(sdk_, (PCO_DescriptionEx*)&pco_desc_3, DESCRIPTION_3);
        if (err)
          throw CameraException(err);
      }

      desc.min_exposure_time_s = static_cast<double>(pco_desc.dwMinExposureDESC) / 1'000'000'000;
      desc.max_exposure_time_s = static_cast<double>(pco_desc.dwMaxExposureDESC) / 1'000;
      desc.min_exposure_step_s = static_cast<double>(pco_desc.dwMinExposureStepDESC) / 1'000'000'000;
      desc.min_delay_time_s = static_cast<double>(pco_desc.dwMinDelayDESC) / 1'000'000'000;
      desc.max_delay_time_s = static_cast<double>(pco_desc.dwMaxDelayDESC) / 1'000;
      desc.min_delay_step_s = static_cast<double>(pco_desc.dwMinDelayStepDESC) / 1'000'000'000;

      if (pco_desc.wIRDESC)
      {
        WORD ir_sensitivity;
        err = PCO_GetIRSensitivity(sdk_, &ir_sensitivity);
        if (err)
          throw CameraException(err);
        if (ir_sensitivity == 1)
        {
          desc.min_exposure_time_s = static_cast<double>(pco_desc.dwMinExposureIRDESC) / 1'000'000'000;
          desc.max_exposure_time_s = static_cast<double>(pco_desc.dwMaxExposureIRDESC) / 1'000;
          desc.min_exposure_step_s = static_cast<double>(pco_desc.dwMinExposureStepDESC) / 1'000'000'000;
          desc.min_delay_time_s = static_cast<double>(pco_desc.dwMinDelayIRDESC) / 1'000'000'000;
          desc.max_delay_time_s = static_cast<double>(pco_desc.dwMaxDelayIRDESC) / 1'000;
          desc.min_delay_step_s = static_cast<double>(pco_desc.dwMinDelayStepDESC) / 1'000'000'000;
        }
      }


      WORD sensor_format;
      err = PCO_GetSensorFormat(sdk_, &sensor_format);
      if (err)
        throw CameraException(err);
      if (sensor_format == SENSORFORMAT_STANDARD)
      {
        desc.max_width = pco_desc.wMaxHorzResStdDESC;
        desc.max_height = pco_desc.wMaxVertResStdDESC;
        if (pco_desc.dwGeneralCapsDESC1 & GENERALCAPS1_ENHANCED_DESCRIPTOR_3)
        {
          desc.min_width = pco_desc_3.wMinHorzResStdDESC3;
          desc.min_height = pco_desc_3.wMinVertResStdDESC3;
        }
        else
        {
          desc.min_width = pco_desc.wMinSizeHorzDESC;
          desc.min_height = pco_desc.wMinSizeVertDESC;
        }
      }
      else
      {
        desc.max_width = pco_desc.wMaxHorzResExtDESC;
        desc.max_height = pco_desc.wMaxVertResExtDESC;
        if (pco_desc.dwGeneralCapsDESC1 & GENERALCAPS1_ENHANCED_DESCRIPTOR_3)
        {
          desc.min_width = pco_desc_3.wMinHorzResExtDESC3;
          desc.min_height = pco_desc_3.wMinVertResExtDESC3;
        }
        else
        {
          desc.min_width = pco_desc.wMinSizeHorzDESC;
          desc.min_height = pco_desc.wMinSizeVertDESC;
        }
      }
      desc.roi_step_horz = pco_desc.wRoiHorStepsDESC;
      desc.roi_step_vert = pco_desc.wRoiVertStepsDESC;

      desc.bit_resolution = pco_desc.wDynResDESC;

      if (pco_desc.dwGeneralCapsDESC1 & GENERALCAPS1_ROI_VERT_SYMM_TO_HORZ_AXIS)
        desc.roi_symmetric_vert = true;

      if (pco_desc.dwGeneralCapsDESC1 & GENERALCAPS1_ROI_HORZ_SYMM_TO_VERT_AXIS)
        desc.roi_symmetric_horz = true;

      desc.has_timestamp_mode = true;
      if (pco_desc.dwGeneralCapsDESC1 & GENERALCAPS1_NO_TIMESTAMP)
        desc.has_timestamp_mode = false;

      desc.has_timestamp_mode_ascii_only = false;
      if (pco_desc.dwGeneralCapsDESC1 & GENERALCAPS1_TIMESTAMP_ASCII_ONLY)
        desc.has_timestamp_mode_ascii_only = true;

      for (int i = 0; i < 4; i++)
      {
        if (pco_desc.dwPixelRateDESC[i] == 0)
          break;

        desc.pixelrate_vec.push_back(pco_desc.dwPixelRateDESC[i]);
      }

      desc.has_acquire_mode = true;
      if (pco_desc.dwGeneralCapsDESC1 & GENERALCAPS1_NO_ACQUIREMODE)
        desc.has_acquire_mode = false;

      desc.has_ext_acquire_mode = false;
      if (pco_desc.dwGeneralCapsDESC1 & GENERALCAPS1_EXT_ACQUIRE)
        desc.has_ext_acquire_mode = true;

      desc.has_metadata_mode = false;
      if (pco_desc.dwGeneralCapsDESC1 & GENERALCAPS1_METADATA)
        desc.has_metadata_mode = true;

      desc.has_ram = true;
      if (pco_desc.dwGeneralCapsDESC1 & GENERALCAPS1_NO_RECORDER)
        desc.has_ram = false;

      WORD current_binning_x = 1;
      while (current_binning_x <= pco_desc.wMaxBinHorzDESC)
      {
        desc.binning_horz_vec.push_back(current_binning_x);
        if (pco_desc.wBinHorzSteppingDESC == 1) //linear
          current_binning_x++;
        else //binary
          current_binning_x *= 2;
      }
      WORD current_binning_y = 1;
      while (current_binning_y <= pco_desc.wMaxBinVertDESC)
      {
        desc.binning_vert_vec.push_back(current_binning_y);
        if (pco_desc.wBinVertSteppingDESC == 1) //linear
          current_binning_y++;
        else //binary
          current_binning_y *= 2;
      }
    }

    return desc;
  }

  void Camera::defaultConfiguration()
  {
    int err = PCO_NOERROR;

    if (hasRam())
    {
      switchToCamRam();
    }

    //______________________________________________________________
    //test if camera is currrently recording and if so then stop
    WORD state;
    err = PCO_GetRecordingState(sdk_, &state);
    if (err)
      throw CameraException(err);
    if (state == 1)
    {
      err = PCO_SetRecordingState(sdk_, 0);
      if (err)
        throw CameraException(err);
    }

    err = PCO_ResetSettingsToDefault(sdk_);
    if (err)
      throw CameraException(err);

    //______________________________________________________________
    //set bit alignment to LSB
    err = PCO_SetBitAlignment(sdk_, BIT_ALIGNMENT_LSB);  //1 = LSB
    if (err)
      throw CameraException(err);

    if (camera_description_.dwGeneralCapsDESC1 & GENERALCAPS1_METADATA)  //then camera can produce metadata
    {
      WORD metadata_size_dummy = 0;
      WORD metadata_version_dummy = 0;
      err = PCO_SetMetaDataMode(sdk_, METADATA_MODE_ON,
        &metadata_size_dummy, &metadata_version_dummy);
      if (err)
        throw CameraException(err);
    }

    //______________________________________________________________
    //load values into camera
    err = PCO_ArmCamera(sdk_);
    if (err)
      throw CameraException(err);
  }

  Configuration Camera::getConfiguration() const
  {
    Configuration config;
    int err = PCO_NOERROR;

    //______________________________________________________________
    //get delay and exposure time
    DWORD exposure_time = UNDEF_DW;
    DWORD delay_time = UNDEF_DW;
    WORD timebase_delay = UNDEF_W;  //0 = ns, 1 = us, 2 = ms
    WORD timebase_exposure = UNDEF_W;
    err = PCO_GetDelayExposureTime(sdk_, &delay_time, &exposure_time, &timebase_delay, &timebase_exposure);
    if (err)
      throw CameraException(err);
    double delay_time_s = static_cast<double>(delay_time) / std::pow(10, 9 - 3 * timebase_delay); //us => delay_time/10^6
    double exposure_time_s = static_cast<double>(exposure_time) / std::pow(10, 9 - 3 * timebase_exposure);

    config.delay_time_s = delay_time_s;
    config.exposure_time_s = exposure_time_s;

    //______________________________________________________________
    //get Hardware-ROI to default (sensor size)
    WORD x0, y0, x1, y1;
    err = PCO_GetROI(sdk_, &x0, &y0, &x1, &y1);
    if (err)
      throw CameraException(err);
    Roi hw_roi{ x0, y0, x1, y1 };
    config.roi = hw_roi;

    //______________________________________________________________
    //get timestamp mode
    WORD timestamp_mode = UNDEF_W;
    if (!(camera_description_.dwGeneralCapsDESC1 & GENERALCAPS1_NO_TIMESTAMP))
    {
      err = PCO_GetTimestampMode(sdk_, &timestamp_mode);
      if (err)
        throw CameraException(err);
    }
    config.timestamp_mode = timestamp_mode;

    //______________________________________________________________
    //get pixel rate
    DWORD pixelrate = UNDEF_DW;
    err = PCO_GetPixelRate(sdk_, &pixelrate);
    if (err)
    {
      if (err & PCO_ERROR_FIRMWARE_NOT_SUPPORTED)
        pixelrate = 0;
      else
        throw CameraException(err);
    }
    config.pixelrate = pixelrate; // PixelRate in Hz

    //______________________________________________________________
    //get trigger mode
    WORD trigger_mode = UNDEF_W;
    err = PCO_GetTriggerMode(sdk_, &trigger_mode);
    if (err)
      throw CameraException(err);
    config.trigger_mode = trigger_mode;

    //______________________________________________________________
    //get acquire mode
    WORD acquire_mode = UNDEF_W;
    if (!(camera_description_.dwGeneralCapsDESC1 & GENERALCAPS1_NO_ACQUIREMODE))
    {
      err = PCO_GetAcquireMode(sdk_, &acquire_mode);
      if (err)
        throw CameraException(err);
    }
    config.acquire_mode = acquire_mode;

    //______________________________________________________________
    //get metadata mode
    if (camera_description_.dwGeneralCapsDESC1 & GENERALCAPS1_METADATA)
    {
      WORD metadata_mode = UNDEF_W;
      WORD metadata_size_dummy = UNDEF_W;
      WORD metadata_version_dummy = UNDEF_W;

      err = PCO_GetMetaDataMode(sdk_, &metadata_mode,
        &metadata_size_dummy, &metadata_version_dummy);
      if (err)
        throw CameraException(err);
      config.metadata_mode = metadata_mode;
    }
    else
      config.metadata_mode = 0x00;
    //______________________________________________________________
    //get noise filter mode
    if (camera_description_.dwGeneralCapsDESC1 & GENERALCAPS1_NOISE_FILTER)
    {
      WORD noise_filter_mode = UNDEF_W;
      err = PCO_GetNoiseFilterMode(sdk_, &noise_filter_mode);
      if (err)
        throw CameraException(err);
      config.noise_filter_mode = noise_filter_mode;
    }


    //______________________________________________________________
    //get binning
    Binning binning;
    binning.horz = BINNING_MASK;
    binning.vert = BINNING_MASK; //necessary, as binning params are ORd
    err = PCO_GetBinning(sdk_, &(binning.horz), &(binning.vert));
    if (err)
      throw CameraException(err);
    config.binning = binning;

    return config;
  }

  void Camera::setConfiguration(Configuration config)
  {
    int err = PCO_NOERROR;
    if (isRecording())
      throw CameraException("Can not set configuration if recoding is running!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_SDKDLL_RECORDER_ACQUISITION_MUST_BE_OFF);
    //______________________________________________________________
    //set delay and exposure time
    DWORD exposure_time = UNDEF_DW;
    WORD timebase_exposure = UNDEF_W; //2=>ms, 1=>us, 0=>ns
    DWORD delay_time = UNDEF_DW;
    WORD timebase_delay = UNDEF_W;

    if (!std::isnan(config.exposure_time_s)
      || !std::isnan(config.delay_time_s))
    {
      if (config.exposure_time_s < 1e-9 * static_cast<double>(camera_description_.dwMinExposureDESC - camera_description_.dwMinExposureStepDESC))
        throw CameraException("invalid exposure time: smaller than min limit or negative", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);
      if (config.exposure_time_s > 1e-3 * static_cast<double>(camera_description_.dwMaxExposureDESC))
        throw CameraException("invalid exposure time: greater than max limit", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);
      if (config.delay_time_s < 1e-9 * static_cast<double>(camera_description_.dwMinDelayDESC))
        throw CameraException("invalid delay time: negative value", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);
      if (config.delay_time_s > 1e-3 * static_cast<double>(camera_description_.dwMaxDelayDESC))
        throw CameraException("invalid delay time: greater than max limit", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);

      if (!std::isnan(config.exposure_time_s))
      {
        calcTimeBase(config.exposure_time_s, &exposure_time, &timebase_exposure);
      }
      else  //use last known exposure time
      {
        DWORD delay_time_dummy;
        WORD timebase_delay_dummy;
        err = PCO_GetDelayExposureTime(sdk_, &delay_time_dummy, &exposure_time,
          &timebase_delay_dummy, &timebase_exposure);
        if (err)
          throw CameraException(err);
      }

      if (!std::isnan(config.delay_time_s))
      {
        calcTimeBase(config.delay_time_s, &delay_time, &timebase_delay);
      }
      else  //use last known delay time
      {
        DWORD exposure_time_dummy;  //PCO_GetDelayExposureTime() doesn't accept nullptr as argument
        WORD timebase_exposure_dummy;
        err = PCO_GetDelayExposureTime(sdk_, &delay_time, &exposure_time_dummy,
          &timebase_delay, &timebase_exposure_dummy);
        if (err)
          throw CameraException(err);
      }

      err = PCO_SetDelayExposureTime(sdk_, delay_time, exposure_time, timebase_delay, timebase_exposure);
      if (err)
        throw CameraException(err);
    }

    //______________________________________________________________
    //set Hardware-ROI
    if (config.roi.x0 != UNDEF_QW
      && config.roi.y0 != UNDEF_QW
      && config.roi.x1 != UNDEF_QW
      && config.roi.y1 != UNDEF_QW)
    {
      err = PCO_SetROI(sdk_,
        static_cast<WORD>(config.roi.x0),
        static_cast<WORD>(config.roi.y0),
        static_cast<WORD>(config.roi.x1),
        static_cast<WORD>(config.roi.y1));
      if (err)
        throw CameraException(err);
    }

    //______________________________________________________________
    //set timestamp_mode
    if (!(camera_description_.dwGeneralCapsDESC1 & GENERALCAPS1_NO_TIMESTAMP))
    {
      if (config.timestamp_mode != UNDEF_W)
      {
        err = PCO_SetTimestampMode(sdk_, config.timestamp_mode);
        if (err)
          throw CameraException(err);
      }
    }

    //______________________________________________________________
    //set pixelrate
    if (config.pixelrate != UNDEF_DW && camera_description_.dwPixelRateDESC[0]) // values from Description.pixelrate_vec, not supported if dwPixelRateDESC[0] == 0
    {
      err = PCO_SetPixelRate(sdk_, config.pixelrate);
      if (err && ((err & PCO_ERROR_FIRMWARE_NOT_SUPPORTED) != PCO_ERROR_FIRMWARE_NOT_SUPPORTED))
        throw CameraException(err);
    }

    //______________________________________________________________
    //set trigger_mode
    if (config.trigger_mode != UNDEF_W)
    {
      err = PCO_SetTriggerMode(sdk_, config.trigger_mode);
      if (err)
        throw CameraException(err);
    }

    //______________________________________________________________
    //set acquire_mode
    if (config.acquire_mode != UNDEF_W)
    {
      if (!(camera_description_.dwGeneralCapsDESC1 & GENERALCAPS1_NO_ACQUIREMODE))
      {
        err = PCO_SetAcquireMode(sdk_, config.acquire_mode);
        if (err)
          throw CameraException(err);
      }
    }

    //______________________________________________________________
    //set metadata_mode
    if (config.metadata_mode != UNDEF_W)
    {
      WORD metadata_size_dummy = 0;
      WORD metadata_version_dummy = 0;
      if (camera_description_.dwGeneralCapsDESC1 & GENERALCAPS1_METADATA)  //then camera can produce metadata
      {
        err = PCO_SetMetaDataMode(sdk_, config.metadata_mode, &metadata_size_dummy, &metadata_version_dummy);
        if (err)
          throw CameraException(err);
      }
    }


    //______________________________________________________________
    //set noise filter mode
    if (config.noise_filter_mode != UNDEF_W)
    {
      if (camera_description_.dwGeneralCapsDESC1 & GENERALCAPS1_NOISE_FILTER)
      {
        err = PCO_SetNoiseFilterMode(sdk_, config.noise_filter_mode);
        if (err)
          throw CameraException(err);
      }
    }

    //______________________________________________________________
    //set binning
    if (config.binning.horz != UNDEF_W
      || config.binning.vert != UNDEF_W)
    {

      WORD b_h = 0;
      WORD b_v = 0;

      if (config.binning.horz != UNDEF_W)
      {
        b_h = config.binning.horz;
      }
      else
      {
        WORD b_v_dummy = 0;
        err = PCO_GetBinning(sdk_, &b_h, &b_v_dummy);
        if (err)
          throw CameraException(err);
      }

      if (config.binning.vert != UNDEF_W)
      {
        b_v = config.binning.vert;
      }
      else
      {
        WORD b_h_dummy = 0;
        err = PCO_GetBinning(sdk_, &b_h_dummy, &b_v);
        if (err)
          throw CameraException(err);
      }

      err = PCO_SetBinning(sdk_, b_h, b_v);
      if (err)
        throw CameraException(err);
    }
    //______________________________________________________________
    //load values into camera
    err = PCO_ArmCamera(sdk_);
    if (err)
      throw CameraException(err);
  }

  void Camera::setExposureTime(double exposure_time_s)
  {
    setDelayExposureTime(getConfiguration().delay_time_s, exposure_time_s);
  }

  double Camera::getExposureTime() //in secs
  {
    return getConfiguration().exposure_time_s;
  }

  void Camera::setDelayTime(double delay_time_s)
  {
    setDelayExposureTime(delay_time_s, getConfiguration().exposure_time_s);
  }

  double Camera::getDelayTime()
  {

    return getConfiguration().delay_time_s;
  }

  //_________________________________________
  // handle recorder

  void Camera::record(DWORD required_recorder_buffer_size, RecordMode record_mode, std::filesystem::path file_path)
  {
    int err = PCO_NOERROR;

    bool blocking = true;

    WORD filemode_no_overwrite = 0; //set to 1, if recored_mode is a file mode, so no images will be overwritten

    WORD memory_type = UNDEF_W;
    WORD memory_mode = 0;



    std::vector<RecordMode> camram_modes = {
      RecordMode::camram_segment,
      RecordMode::camram_ring }; // just required for live view
    if (std::find(camram_modes.begin(), camram_modes.end(), record_mode) != camram_modes.end())
    {
      if (!hasRam())
        throw CameraException("Camera does not support CamRam!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);

      DWORD max_images = getCamRamMaxImages();
      if (!max_images)
        throw CameraException("Camera has no memory allocated for the active segment", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_BUFFERSIZE);

      if (required_recorder_buffer_size != max_images)
        std::cout << "CamRam record mode: required_recorder_buffer_size is set automatically to maximum : " << max_images << "\n";
      required_recorder_buffer_size = max_images;

      //______________________________________________________________
      //turn metadata_mode always on for camram
      WORD metadata_size_dummy = 0;
      WORD metadata_version_dummy = 0;
      if (camera_description_.dwGeneralCapsDESC1 & GENERALCAPS1_METADATA)  //then camera can produce metadata
      {
        err = PCO_SetMetaDataMode(sdk_, METADATA_MODE_ON, &metadata_size_dummy, &metadata_version_dummy);
        if (err)
          throw CameraException(err);
        
        metadata_mode_on_ = true;
      }

      err = PCO_SetStorageMode(sdk_, STORAGE_MODE_RECORDER);
      if (err)
        throw CameraException(err);

      if (record_mode == RecordMode::camram_segment)
      {
        err = PCO_SetRecorderSubmode(sdk_, 0); //switch to sequence mode
        if (err)
          throw CameraException(err);
      }
      else
      {
        err = PCO_SetRecorderSubmode(sdk_, 1); // default: ring buffer mode
        if (err)
          throw CameraException(err);
      }
    }

    if (!required_recorder_buffer_size)
      throw CameraException(PCO_ERROR_WRONGVALUE);

    err = PCO_ArmCamera(sdk_);  //historic, see documantation
    if (err)
      throw CameraException(err); // TODO: test!, read bit if armed

    //______________________________________________________________
    //see if timestamp mode contains binary. if not, image() must not ask for it
    WORD timestamp_mode = UNDEF_W;
    err = PCO_GetTimestampMode(sdk_, &timestamp_mode);
    if (err)
      throw CameraException(err);
    if (timestamp_mode == TIMESTAMP_MODE_BINARY
      || timestamp_mode == TIMESTAMP_MODE_BINARYANDASCII)
      timestamp_mode_binary_ = true;
    else
      timestamp_mode_binary_ = false;

    //______________________________________________________________
    //see if metadata mode is activated. if not, image() must not ask for it
    WORD metadata_mode = UNDEF_W;
    WORD metadata_size_dummy = UNDEF_W;
    WORD metadata_version_dummy = UNDEF_W;

    if ((camera_description_.dwGeneralCapsDESC1 & GENERALCAPS1_METADATA))
    {
      err = PCO_GetMetaDataMode(sdk_, &metadata_mode,
        &metadata_size_dummy, &metadata_version_dummy);

      if (err)
        throw CameraException(err);
      metadata_mode_on_ = static_cast<bool>(metadata_mode);
    }
    else
      metadata_mode_on_ = false;

    //______________________________________________________________
    //get Hardware-ROI to default (sensor size)
    WORD x0, y0, x1, y1;
    err = PCO_GetROI(sdk_, &x0, &y0, &x1, &y1);

    if (err)
      throw CameraException(err);
    hard_roi_.x0 = x0;
    hard_roi_.y0 = y0;
    hard_roi_.x1 = x1;
    hard_roi_.y1 = y1;
    //______________________________________________________________
    //see, if double shutter is activated. If so then twice the memory is required
    if (camera_description_.wDoubleImageDESC)  //does camera support double shutter?
    {
      WORD double_image = UNDEF_W;
      err = PCO_GetDoubleImageMode(sdk_, &double_image);
      if (err)
        throw CameraException(err);

      if (double_image)  //is double shutter activated?
        double_image_factor_ = 2;
      else
        double_image_factor_ = 1;
    }

    //______________________________________________________________
    //create recorder
    //___________________________________________
    //delete if already exists
    if (rec_ != nullptr)
    {
      err = PCO_RecorderStopRecord(rec_, sdk_);
      if (err)
        throw CameraException(err);
      err = PCO_RecorderDelete(rec_);
      if (err)
        throw CameraException(err);
      rec_ = nullptr;
    }

    //________________________________________________________
    // update all convert handles to sensor info that may have changed
    PCO_SensorInfo sensor_info = getSensorInfo();
    for (auto& [format, handle] : conv_map_)
    {
      err = PCO_ConvertSetSensorInfo(conv_map_[format], &sensor_info);
      if (err)
        throw CameraException(err);
    }

    // default volume
    std::string root_volume;
    if (!file_path.parent_path().empty())
      root_volume = file_path.parent_path().string();
    else
      root_volume = file_path.string();

    // camram off
    WORD segment = 0;

    //___________________________________________
    //create recorder according to specified mode
    switch (record_mode)
    {
      //__________________
      //basic modes
    case RecordMode::sequence_dpcore:
    {
      memory_mode |= PCO_RECORDER_MODE_USE_DPCORE;
    }
    [[fallthrough]];
    case RecordMode::sequence:
    {
      memory_type = PCO_RECORDER_MEMORY_SEQUENCE;
      memory_mode |= PCO_RECORDER_MODE_MEMORY;

      if (!file_path.empty())
        throw CameraException(PCO_ERROR_FILE_WRONG_VALUE);
      break;
    }

    case RecordMode::sequence_non_blocking_dpcore:
    {
      memory_mode |= PCO_RECORDER_MODE_USE_DPCORE;
    }
    [[fallthrough]];
    case RecordMode::sequence_non_blocking:
    {
      blocking = false;

      memory_type = PCO_RECORDER_MEMORY_SEQUENCE;
      memory_mode |= PCO_RECORDER_MODE_MEMORY;

      if (!file_path.empty())
        throw CameraException(PCO_ERROR_FILE_WRONG_VALUE);

      break;
    }

    case RecordMode::ring_buffer_dpcore:
    {
      memory_mode |= PCO_RECORDER_MODE_USE_DPCORE;
    }
    [[fallthrough]];
    case RecordMode::ring_buffer:
    {
      blocking = false;

      memory_type = PCO_RECORDER_MEMORY_RINGBUF;
      memory_mode |= PCO_RECORDER_MODE_MEMORY;

      if (required_recorder_buffer_size < 4)
        throw CameraException(PCO_ERROR_BUFFERSIZE);

      if (!file_path.empty())
        throw CameraException(PCO_ERROR_FILE_WRONG_VALUE);

      break;
    }

    case RecordMode::fifo_dpcore:
    {
      memory_mode |= PCO_RECORDER_MODE_USE_DPCORE;
    }
    [[fallthrough]];
    case RecordMode::fifo:
    {
      blocking = false;
      memory_type = PCO_RECORDER_MEMORY_FIFO;
      memory_mode |= PCO_RECORDER_MODE_MEMORY;

      if (required_recorder_buffer_size < 4)
        throw CameraException(PCO_ERROR_BUFFERSIZE);

      if (!file_path.empty())
        throw CameraException(PCO_ERROR_FILE_WRONG_VALUE);

      break;
    }

    //__________________
    //file modes
    case RecordMode::tif:
    {
      blocking = false;
      memory_type = PCO_RECORDER_FILE_TIF;
      memory_mode = PCO_RECORDER_MODE_FILE;
      filemode_no_overwrite = 1;

      if (file_path.empty())
        throw CameraException(PCO_ERROR_FILE_WRONG_VALUE);
      break;
    }
    case RecordMode::multitif:
    {
      blocking = false;
      memory_type = PCO_RECORDER_FILE_MULTITIF;
      memory_mode = PCO_RECORDER_MODE_FILE;
      filemode_no_overwrite = 1;

      if (file_path.empty())
        throw CameraException(PCO_ERROR_FILE_WRONG_VALUE);
      break;
    }
    case RecordMode::pcoraw:
    {
      blocking = false;
      memory_type = PCO_RECORDER_FILE_PCORAW;
      memory_mode = PCO_RECORDER_MODE_FILE;
      filemode_no_overwrite = 1;

      if (file_path.empty())
        throw CameraException(PCO_ERROR_FILE_WRONG_VALUE);
      break;
    }
    case RecordMode::b16:
    {
      blocking = false;
      memory_type = PCO_RECORDER_FILE_B16;
      memory_mode = PCO_RECORDER_MODE_FILE;
      filemode_no_overwrite = 1;

      if (file_path.empty())
        throw CameraException(PCO_ERROR_FILE_WRONG_VALUE);
      break;
    }
    case RecordMode::dicom:
    {
      blocking = false;
      memory_type = PCO_RECORDER_FILE_DICOM;
      memory_mode = PCO_RECORDER_MODE_FILE;
      filemode_no_overwrite = 1;

      if (file_path.empty())
        throw CameraException(PCO_ERROR_FILE_WRONG_VALUE);
      break;
    }
    case RecordMode::multidicom:
    {
      blocking = false;
      memory_type = PCO_RECORDER_FILE_MULTIDICOM;
      memory_mode = PCO_RECORDER_MODE_FILE;
      filemode_no_overwrite = 1;

      if (file_path.empty())
        throw CameraException(PCO_ERROR_FILE_WRONG_VALUE);

      break;
    }
    //__________________
    // camram modes
    case RecordMode::camram_ring:
    {}
    [[fallthrough]];
    case RecordMode::camram_segment:
    {
      blocking = false;
      memory_mode = PCO_RECORDER_MODE_CAMRAM;
      memory_type = PCO_RECORDER_CAMRAM_SEQUENTIAL;

      segment = getCamRamSegment();

      if (!file_path.empty())
        throw CameraException(PCO_ERROR_FILE_WRONG_VALUE);
      break;
    }

    //__________________
    default:
    {
      throw CameraException(PCO_ERROR_WRONGVALUE);
      break;
    }
    }

    if (root_volume.empty())
      err = PCO_RecorderCreate(&rec_, &sdk_, nullptr, 1, memory_mode, nullptr, &recorder_buffer_size_);
    else
      err = PCO_RecorderCreate(&rec_, &sdk_, nullptr, 1, memory_mode, root_volume.c_str(), &recorder_buffer_size_);
    if (err)
    {
      rec_ = nullptr;
      throw CameraException(err);
    }

    //___________________________________________
    //see if recorder is able to store required images and cut accordingly
    if (required_recorder_buffer_size > recorder_buffer_size_)
    {
      required_recorder_buffer_size = recorder_buffer_size_;
      std::cout << "Not enough space for your application." << std::endl;
      std::cout << "Reqired number of images get adapted to max possible." << std::endl;
    }
    else if (required_recorder_buffer_size > recorder_buffer_size_ / 2 && !segment)
    {
      std::cout << "You are above 50% of available space." << std::endl;
    }

    //___________________________________________
    //initialize recorder and start record

    if (segment)
    {
      err = PCO_RecorderInit(rec_, &required_recorder_buffer_size,
        1, //only one camera allowed
        memory_type, 0, nullptr, &segment);
    }
    else
    {
      err = PCO_RecorderInit(rec_, &required_recorder_buffer_size,
        1, //only one camera allowed
        memory_type, filemode_no_overwrite, file_path.string().c_str(), nullptr);
    }

    if (err)
    {
      // ~Camera eventually deletes recorder
      throw CameraException(err);
    }
    err = PCO_RecorderStartRecord(rec_, sdk_);
    if (err)
      throw CameraException(err);

    for (auto& [data_format, reset] : do_auto_minmax_map_)
      reset = true;

    //___________________________________________
    //wait until images are recorded
    if (blocking)
    {
      while (1)
      {
        bool is_running = false;
        err = PCO_RecorderGetStatus(rec_, sdk_, &is_running, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr);
        if (err)
          throw CameraException(err);
        if (!is_running)
          break;
        std::this_thread::sleep_for(std::chrono::milliseconds(WAIT_TIME_MS));
      }
    }
  }

  void Camera::stop()
  {
    int err = PCO_NOERROR;
    current_image_number_ = 0;
    if (rec_ != nullptr)
    {
      err = PCO_RecorderStopRecord(rec_, sdk_);
      if (err)
        throw CameraException(err);
    }
  }

  void Camera::waitForFirstImage(bool delay, double timeout_s)
  {
    if (!rec_)
      throw CameraException("Camera has no recorder yet. Call \"record(...)\" before \"waitForFirstImage(...)\"", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_SDKDLL_RECORDER_NO_IMAGES_AVAILABLE);
    
    DWORD image_count = UNDEF_DW;
    const std::clock_t start_time = std::clock();
    double duration = 0.0;

    int err = PCO_NOERROR;
    while (1)
    {
      err = PCO_RecorderGetStatus(rec_, sdk_, nullptr, nullptr, nullptr,
        &image_count,     //get index of last recorded image
        nullptr, nullptr, nullptr, nullptr, nullptr);
      if (err)
        throw CameraException(err);

      if (image_count >= 1)
        break;
      if (delay)
        std::this_thread::sleep_for(std::chrono::milliseconds(WAIT_TIME_MS));

      duration = ((double)std::clock() - (double)start_time) / CLOCKS_PER_SEC;
      if (!std::isnan(timeout_s) && (duration >= timeout_s))
      {
        err = PCO_ERROR_APPLICATION_PICTURETIMEOUT;
        throw CameraException(err);
      }
    }
  }

  void Camera::waitForNewImage(bool delay, double timeout_s)
  {
    if (!rec_)
      throw CameraException("Camera has no recorder yet. Call \"record(...)\" before \"waitForNewImage(...)\"", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_SDKDLL_RECORDER_NO_IMAGES_AVAILABLE);
    
    DWORD image_count = UNDEF_DW;
    const std::clock_t start_time = std::clock();
    double duration = 0.0;
    int err = PCO_NOERROR;
    bool is_running = true;
    DWORD mode_and_type = 0;
    err = PCO_RecorderGetSettings(rec_, sdk_, &mode_and_type, nullptr, nullptr, nullptr, nullptr, nullptr);
    if (err)
      throw CameraException(err);

    while (1)
    {
      err = PCO_RecorderGetStatus(rec_, sdk_, &is_running, nullptr, nullptr,
        &image_count,     //get index of last recorded image
        nullptr, nullptr, nullptr, nullptr, nullptr);
      if (err)
        throw CameraException(err);

      if (!is_running)
        break;

      if ((HIWORD(mode_and_type) & PCO_RECORDER_MODE_CAMRAM) == PCO_RECORDER_MODE_CAMRAM
        && current_image_number_ == UINT_MAX) // no metadata with img counter available: UINT_MAX
      {
        break;
      }

      if (((LOWORD(mode_and_type) & PCO_RECORDER_MEMORY_FIFO) == PCO_RECORDER_MEMORY_FIFO) && image_count > 0)
        break;
      if (image_count > current_image_number_)
        break;
      if (delay)
        std::this_thread::sleep_for(std::chrono::milliseconds(WAIT_TIME_MS));

      duration = ((double)std::clock() - (double)start_time) / CLOCKS_PER_SEC;
      if (!std::isnan(timeout_s) && (duration >= timeout_s))
      {
        err = PCO_ERROR_APPLICATION_PICTURETIMEOUT;
        throw CameraException(err);
      }
    }
  }

  size_t Camera::getRecordedImageCount() const
  {
    if (!rec_)
      return 0;
    int err = PCO_NOERROR;
    DWORD recorded_image_count = UNDEF_DW;
    err = PCO_RecorderGetStatus(rec_, sdk_, nullptr, nullptr, nullptr, &recorded_image_count, nullptr, nullptr, nullptr, nullptr, nullptr);
    if (err)
      throw CameraException(err);
    return static_cast<size_t>(recorded_image_count);
  }

  size_t Camera::getRecorderBufferSize() const
  {
    if (!rec_)
      return 0;
    int err = PCO_NOERROR;
    DWORD recorder_buffer_size = UNDEF_DW;
    err = PCO_RecorderGetStatus(rec_, sdk_, nullptr, nullptr, nullptr, nullptr, &recorder_buffer_size, nullptr, nullptr, nullptr, nullptr);
    if (err)
      throw CameraException(err);
    return static_cast<size_t>(recorder_buffer_size);
  }

  //_________________________________________
  // convert

  ConvertControl Camera::getConvertControl(DataFormat data_format)
  {
    if (conv_map_.find(data_format) == conv_map_.end()) // not exist
    {
      throw CameraException("Data format is not supported by the camera!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);
    }
    if (conv_ctrl_map_.find(data_format) == conv_ctrl_map_.end())
    {
      throw CameraException("DataFormat is not valid for this camera type!");
    }

    PCO_Display display{ sizeof(PCO_Display) };
    DWORD err = PCO_ConvertGetDisplay(conv_map_.at(data_format), &display);
    if (err)
      throw CameraException(err);

    std::visit(overloaded{
      [&display](ConvertControlMono& cc_mono) {
          cc_mono.gamma = static_cast<double>(display.iGamma) / 100.0;
          cc_mono.contrast = display.iContrast;
          cc_mono.min_limit = display.iScale_min;
          cc_mono.max_limit = display.iScale_max;
        },
      [&display](ConvertControlColor& cc_color) {
          cc_color.gamma = static_cast<double>(display.iGamma) / 100.0;
          cc_color.contrast = display.iContrast;
          cc_color.min_limit = display.iScale_min;
          cc_color.max_limit = display.iScale_max;
          cc_color.color_temperature = display.iColor_temp;
          cc_color.color_saturation = display.iColor_saturation;
          cc_color.color_vibrance = display.iColor_vibrance;
          cc_color.color_tint = display.iColor_tint;
        },
      [&display](ConvertControlPseudoColor& cc_color) {
          cc_color.gamma = static_cast<double>(display.iGamma) / 100.0;
          cc_color.contrast = display.iContrast;
          cc_color.min_limit = display.iScale_min;
          cc_color.max_limit = display.iScale_max;
          cc_color.color_temperature = display.iColor_temp;
          cc_color.color_saturation = display.iColor_saturation;
          cc_color.color_vibrance = display.iColor_vibrance;
          cc_color.color_tint = display.iColor_tint;
        }
      }, conv_ctrl_map_.at(data_format));
    return conv_ctrl_map_.at(data_format);
  }

  void Camera::setConvertControl(DataFormat data_format, ConvertControl variant_settings)
  {
    if (conv_ctrl_map_.find(data_format) == conv_ctrl_map_.end())
      throw CameraException("DataFormat is not valid for this camera type!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);
    if (conv_ctrl_map_.at(data_format).index() != variant_settings.index())
      throw CameraException("ConvertControl struct does not fit to the DataFormat!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);

    if (std::holds_alternative<ConvertControlPseudoColor>(variant_settings))
    {
      auto& lp = std::get<ConvertControlPseudoColor>(variant_settings).lut_file;
      if (!lp.empty())
      {
        if (!std::filesystem::exists(lp))
          throw CameraException("LUT file does not exist!", PCO_ERROR_SDKAPPLICATION | PCO_ERROR_FILE_OPEN);

        int lut_format = 0;
        switch (lp.extension().string()[3])
        {
        case '1':
          lut_format = 3;
          break;
        case '2':
          lut_format = 2;
          break;
        case '3':
          lut_format = 1;
          break;
        case '4':
          lut_format = 0;
          break;
        default: // or .lut, ...
          lut_format = 1;
          break;
        }

        auto s_path = lp.string();
        std::vector<char> c_path(s_path.length() + 1);
        std::copy(s_path.begin(), s_path.end(), c_path.begin());
        int err = PCO_LoadPseudoLut(conv_map_[data_format], lut_format, c_path.data());
        if (err)
          throw CameraException(err);
      }
    }
    conv_ctrl_map_[data_format] = variant_settings;
    checkMinMaxScale(data_format);
    updateDisplaySettings(data_format);
  }

  void Camera::loadLut(DataFormat data_format, std::filesystem::path lut_file)
  {
    if (isColored())
    {
      throw CameraException("Pseudo color not supported for color cameras!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_APPLICATION_FUNCTIONNOTSUPPORTED);
    }

    ConvertControl conv_ctrl = getConvertControl(data_format);
    if (conv_ctrl_map_.find(data_format) == conv_ctrl_map_.end())
      throw CameraException("DataFormat is not valid for this camera type!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);
    if (!std::holds_alternative<ConvertControlPseudoColor>(conv_ctrl))
      throw CameraException("ConvertControlPseudoColor is not assigned to specified DataFormat!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);

    ConvertControlPseudoColor cc = std::get<ConvertControlPseudoColor>(conv_ctrl);
    cc.lut_file = lut_file;
    setConvertControl(data_format, cc);
  }

  void Camera::adaptWhiteBalance(Image& image)
  {
    size_t size_x = image.width();
    size_t size_y = image.height();

    Roi full_size{ 1, 1, size_x, size_y };
    adaptWhiteBalance(image, full_size);
  }

  void Camera::adaptWhiteBalance(Image& image, Roi roi)
  {
    DataFormat data_format = image.getDataFormat();
    if (conv_map_.find(data_format) == conv_map_.end()) // not exist
    {
      throw CameraException("Data format is not supported by the camera!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);
    }
    if (conv_ctrl_map_.find(data_format) == conv_ctrl_map_.end()) // assert
    {
      throw CameraException("ConvertControl not available for this DataFormat!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);
    }

    if (!image.isColored())
    {
      throw CameraException("White balance adaptation is supported for colored data formats!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_APPLICATION_FUNCTIONNOTSUPPORTED);
    }
    if ((roi.width() > image.width()) || (roi.height() > image.height()))
    {
      throw CameraException("Roi is too large", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);
    }

    int color_temp = 0;
    int tint = 0;
    int err = PCO_GetWhiteBalance(conv_map_.at(data_format), &color_temp, &tint, 0,
      (int)roi.width(), (int)roi.height(),
      image.raw_vector_16bit().data(),
      (int)roi.x0, (int)roi.y0, (int)roi.x1, (int)roi.y1);
    if (err)
      throw CameraException(err);

    std::visit(overloaded{
      [&color_temp, &tint](ConvertControlColor& cc_color) {
          cc_color.color_temperature = color_temp;
          cc_color.color_tint = tint;

        },
      [&color_temp, &tint](ConvertControlPseudoColor& cc_color) {
          cc_color.color_temperature = color_temp;
          cc_color.color_tint = tint;
        },
      [](auto&) {
          //pass
        }
      }, conv_ctrl_map_.at(data_format));

    checkMinMaxScale(data_format);
    updateDisplaySettings(data_format);
  }

  PCO_SensorInfo Camera::getSensorInfo() const
  {
    PCO_SensorInfo sensor_info_{ sizeof(PCO_SensorInfo) };
    //______________________________________________________________
    //set internal sensor values and transfer later to image class
    WORD bit_alignment = UNDEF_W;

    int err = PCO_GetBitAlignment(sdk_, &bit_alignment);
    if (err)
      throw CameraException(err);

    if (bit_alignment == BIT_ALIGNMENT_MSB)
      sensor_info_.iSensorInfoBits |= 0x0002;

    //Is color or not
    if (camera_description_.wSensorTypeDESC & 0x0001)
    {
      sensor_info_.iSensorInfoBits |= 0x01;
    }

    double ccm[9] = { 1, 0, 0,
                      0, 1, 0,
                      0, 0, 1 };

    if (camera_description_.dwGeneralCapsDESC1 & GENERALCAPS1_CCM)  //if Camera has CCM, let ccm be set automatically
    {
      err = PCO_GetColorCorrectionMatrix(sdk_, ccm);
      if (err)
        throw CameraException(err);
    }

    sensor_info_.iConversionFactor = 0;
    sensor_info_.iDataBits = camera_description_.wDynResDESC;
    WORD dark_offset = 100;
    err = PCO_GetSensorDarkOffset(sdk_, &dark_offset);
    if (err)
      throw CameraException(err);
    sensor_info_.iDarkOffset = static_cast<int>(dark_offset);

    sensor_info_.strColorCoeff = { ccm[0], ccm[1], ccm[2],
                                  ccm[3], ccm[4], ccm[5],
                                  ccm[6], ccm[7], ccm[8] };

    sensor_info_.iCamNum = 0;
    sensor_info_.hCamera = sdk_;

    return sensor_info_;
  }

  //_________________________________________
  // images

  void Camera::image(Image& image_ref, DWORD image_index, DataFormat data_format, PCO_Recorder_CompressionParams* comp_params)//, ConvertControl* convert_params)
  {
    size_t size_x = hard_roi_.x1 - hard_roi_.x0 + 1;
    size_t size_y = (hard_roi_.y1 - hard_roi_.y0 + 1) * double_image_factor_;

    Roi full_size{ 1, 1, size_x, size_y };
    image(image_ref, full_size, image_index, data_format, comp_params);
  }

  void Camera::image(Image& image_ref, Roi roi, DWORD image_index, DataFormat data_format, PCO_Recorder_CompressionParams* comp_params)
  {
    if (!rec_)
    {
      if (hasRam())
        throw CameraException("Camera has no memory allocated for the active segment", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_BUFFERSIZE);
      throw CameraException("Camera has not recorded yet. Call \"record(...)\" before \"image(...)\"", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_SDKDLL_RECORDER_NO_IMAGES_AVAILABLE);
    }

    if (!isColored())
    {
      if (data_format == DataFormat::BGR16)
        throw CameraException("DataFormat::BGR16 is not supported for mono cameras!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);
    }

    int err = PCO_NOERROR;
    if (data_format == DataFormat::CompressedMono8)
    {
      if (!comp_params)
        throw CameraException("Compression parameters are required for DataFormat::CompressedMono8!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);
      else
        err = PCO_RecorderSetCompressionParams(rec_, sdk_, comp_params);
      if (err)
        throw CameraException(err);
    }

    if (!image_ref.validAllocation(roi, data_format)) // raw + conversion output data
      image_ref.resize(roi, data_format);

    //do not ask for metadata/timestamp if they are turned off
    PCO_METADATA_STRUCT* meta_data_ptr = nullptr;
    PCO_TIMESTAMP_STRUCT* timestamp_ptr = nullptr;
    if (metadata_mode_on_) // || (image_index == PCO_RECORDER_LATEST_IMAGE && recIsCamRam()))
    {
      if (!image_ref.getMetaDataPtr())
        image_ref.setMetaData();
      meta_data_ptr = image_ref.getMetaDataPtr();
    }

    if (timestamp_mode_binary_)
    {
      if (!image_ref.getTimestampPtr())
        image_ref.setTimestamp();
      timestamp_ptr = image_ref.getTimestampPtr();
    }

    if (data_format == DataFormat::CompressedMono8)
    {
      err = PCO_RecorderCopyImageCompressed(rec_, sdk_, image_index,
        static_cast<WORD>(roi.x0),
        static_cast<WORD>(roi.y0),
        static_cast<WORD>(roi.x1),
        static_cast<WORD>(roi.y1),
        static_cast<BYTE*>(image_ref.data().first), &current_image_number_,
        meta_data_ptr, timestamp_ptr);
      if (err)
        throw CameraException(err);

      image_ref.setRecorderImageNumber(current_image_number_);
      return;
    }

    err = PCO_RecorderCopyImage(rec_, sdk_, image_index,
      static_cast<WORD>(roi.x0),
      static_cast<WORD>(roi.y0),
      static_cast<WORD>(roi.x1),
      static_cast<WORD>(roi.y1),
      static_cast<WORD*>(image_ref.raw_data().first), &current_image_number_,
      meta_data_ptr, timestamp_ptr);
    if (err)
      throw CameraException(err);

    image_ref.setRecorderImageNumber(current_image_number_);



    if (data_format != DataFormat::Mono16) // TODO: extend for 8bit raw data
    {
      if (conv_map_.find(data_format) == conv_map_.end()) // not exist
      {
        throw CameraException("Data format is not supported by the camera!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);
      }
      if (conv_ctrl_map_.find(data_format) == conv_ctrl_map_.end()) // assert
      {
        throw CameraException("ConvertControl not available for this DataFormat!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);
      }

      int mode = CONVERT_MODE_EXT_FILTER_FLAGS;

      std::visit(overloaded{
        [&mode](ConvertControlColor& cc) {
          if (cc.sharpen)
            mode |= CONVERT_MODE_OUT_DOSHARPEN;
          if (cc.adaptive_sharpen)
            mode |= CONVERT_MODE_OUT_DOADSHARPEN;
          if (cc.flip_vertical)
            mode |= CONVERT_MODE_OUT_FLIPIMAGE;
          if (cc.pco_debayer_algorithm)
            mode |= CONVERT_MODE_OUT_DOPCODEBAYER;
          },
        [&mode](auto& cc) {
          if (cc.sharpen)
            mode |= CONVERT_MODE_OUT_DOSHARPEN;
          if (cc.adaptive_sharpen)
            mode |= CONVERT_MODE_OUT_DOADSHARPEN;
          if (cc.flip_vertical)
            mode |= CONVERT_MODE_OUT_FLIPIMAGE;
          }
        }, conv_ctrl_map_.at(data_format));

      bool auto_minmax = std::visit([](auto& cc) {return cc.auto_minmax; }, conv_ctrl_map_.at(data_format));
      if (auto_minmax)
      {
        mode |= CONVERT_MODE_OUT_AUTOMINMAX;
        if (do_auto_minmax_map_.at(data_format))
        {
          setAutoMinMax(image_ref, data_format);
          do_auto_minmax_map_[data_format] = false;
        }
        if (checkMinMaxScale(data_format)) //check if we need to limit boundaries
          updateDisplaySettings(data_format);
      }

      if (data_format == DataFormat::BGRA8)
        mode |= CONVERT_MODE_OUT_RGB32;

      if (data_format == DataFormat::Mono8)
      {
        err = PCO_Convert16TO8(conv_map_[data_format], mode, calculateColorMode(roi),
          static_cast<int>(roi.width()), static_cast<int>(roi.height()),
          static_cast<WORD*>(image_ref.raw_data().first), static_cast<BYTE*>(image_ref.data().first));
      }
      else if (data_format == DataFormat::BGR16)
      {
        err = PCO_Convert16TOCOL16(conv_map_[data_format], mode, calculateColorMode(roi),
          static_cast<int>(roi.width()), static_cast<int>(roi.height()),
          static_cast<WORD*>(image_ref.raw_data().first), static_cast<WORD*>(image_ref.data().first));
      }
      else
      {
        if (isColored())
        {
          err = PCO_Convert16TOCOL(conv_map_[data_format], mode, calculateColorMode(roi),
            static_cast<int>(roi.width()), static_cast<int>(roi.height()),
            static_cast<WORD*>(image_ref.raw_data().first), static_cast<BYTE*>(image_ref.data().first));
        }
        else
        {
          err = PCO_Convert16TOPSEUDO(conv_map_[data_format], mode, calculateColorMode(roi),
            static_cast<int>(roi.width()), static_cast<int>(roi.height()),
            static_cast<WORD*>(image_ref.raw_data().first), static_cast<BYTE*>(image_ref.data().first));
        }
      }
      if (err)
        throw CameraException(err);
    }
  }

  void Camera::images(std::vector<Image>& image_sequence, DataFormat data_format, size_t start_index, PCO_Recorder_CompressionParams* comp_params)
  {
    size_t size_x = hard_roi_.x1 - hard_roi_.x0 + 1;
    size_t size_y = (hard_roi_.y1 - hard_roi_.y0 + 1) * double_image_factor_;

    Roi full_size{ 1, 1, size_x, size_y };
    images(image_sequence, full_size, data_format, start_index, comp_params);
  }

  void Camera::images(std::vector<Image>& image_sequence, Roi roi, DataFormat data_format, size_t start_index, PCO_Recorder_CompressionParams* comp_params)
  {
    DWORD recorded_image_count = UNDEF_DW;

    int err = PCO_NOERROR;

    if (!rec_)
    {
      if (hasRam())
        throw CameraException("Camera has no memory allocated for the active segment", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_BUFFERSIZE);
      throw CameraException("Camera has not recorded yet. Call \"record(...)\" before \"images(...)\"", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_SDKDLL_RECORDER_NO_IMAGES_AVAILABLE);
    }

    // pre-allocate
    for (auto& img : image_sequence)
    {
      if (!img.validAllocation(roi, data_format))
        img.resize(roi, data_format);
    }

    //___________________________________________
    //see if required number of images are recorded / can be recorded
    DWORD req_image_count = 0;
    err = PCO_RecorderGetStatus(rec_, sdk_, nullptr, nullptr, nullptr, &recorded_image_count, &req_image_count, nullptr, nullptr, nullptr, nullptr);
    if (err)
      throw CameraException(err);

    if (!image_sequence.size()) // unspecified vector
    {
      image_sequence.resize(req_image_count - start_index, Image(roi, data_format, RawFormat::Word));
    }
    if (image_sequence.size() > req_image_count - start_index)
    {
      throw CameraException("Image vector is too big to hold latest recorded images starting from start_index!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);
    }

    while (1) //wait until required number of images are recorded
    {
      err = PCO_RecorderGetStatus(rec_, sdk_, nullptr, nullptr, nullptr, &recorded_image_count, nullptr, nullptr, nullptr, nullptr, nullptr);
      if (err)
        throw CameraException(err);
      if (recorded_image_count >= (image_sequence.size() + start_index))
        break;
    }

    //___________________________________________
    //copy all required images from recorder into vector
    for (auto& img : image_sequence)
    {
      image(img, roi, static_cast<DWORD>(start_index++), data_format, comp_params);
    }
  }

  void Camera::imageAverage(Image& image_ref, DataFormat data_format)
  {
    size_t size_x = hard_roi_.x1 - hard_roi_.x0 + 1;
    size_t size_y = (hard_roi_.y1 - hard_roi_.y0 + 1) * double_image_factor_;

    Roi full_size{ 1, 1, size_x, size_y };
    return imageAverage(image_ref, full_size, data_format);
  }

  void Camera::imageAverage(Image& image_ref, Roi roi, DataFormat data_format)
  {
    if (!rec_)
    {
      if (hasRam())
        throw CameraException("Camera has no memory allocated for the active segment", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_BUFFERSIZE);
      throw CameraException("Camera has not recorded yet. Call \"record(...)\" before \"imageAverage(...)\"", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_SDKDLL_RECORDER_NO_IMAGES_AVAILABLE);
    }

    if (data_format == DataFormat::CompressedMono8)
      throw CameraException("DataFormat::CompressedMono8 is not supported for average images!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);

    if (!isColored())
    {
      if (data_format == DataFormat::BGR16)
        throw CameraException("DataFormat::BGR16 is not supported for mono cameras!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);
    }

    int err = PCO_NOERROR;
    if (!image_ref.validAllocation(roi, data_format)) // raw + conversion output data
      image_ref.resize(roi, data_format);

    // do not exist for average images
    image_ref.clearMetaData();
    image_ref.clearTimestamp();


    DWORD start = 0;
    //___________________________________________
    //average of all recorded images => get count
    DWORD recorded_image_count;
    DWORD recquired_image_count;
    err = PCO_RecorderGetStatus(rec_, sdk_, nullptr, nullptr, nullptr, &recorded_image_count,
      &recquired_image_count, nullptr, nullptr, nullptr, nullptr);
    if (err)
      throw CameraException(err);

    DWORD stop_index = std::min(recorded_image_count, recquired_image_count) - 1; //zero based idx

    //___________________________________________
    //copy image from recorder
    err = PCO_RecorderCopyAverageImage(rec_, sdk_, start, stop_index,
      static_cast<WORD>(roi.x0),
      static_cast<WORD>(roi.y0),
      static_cast<WORD>(roi.x1),
      static_cast<WORD>(roi.y1),
      static_cast<WORD*>(image_ref.raw_data().first)); // TODO: 8bit not implemented by recorder
    image_ref.setRecorderImageNumber(stop_index);
    if (err)
      throw CameraException(err);

    if (data_format != DataFormat::Mono16) // TODO: extend for 8bit raw data. Not implemented by convert
    {
      if (conv_map_.find(data_format) == conv_map_.end()) // not exist
      {
        throw CameraException("Data format is not supported by the camera!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);
      }
      if (conv_ctrl_map_.find(data_format) == conv_ctrl_map_.end()) // assert
      {
        throw CameraException("ConvertControl not available for this DataFormat!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);
      }

      int mode = CONVERT_MODE_EXT_FILTER_FLAGS;

      std::visit(overloaded{
        [&mode](ConvertControlColor& cc) {
          if (cc.sharpen)
            mode |= CONVERT_MODE_OUT_DOSHARPEN;
          if (cc.adaptive_sharpen)
            mode |= CONVERT_MODE_OUT_DOADSHARPEN;
          if (cc.flip_vertical)
            mode |= CONVERT_MODE_OUT_FLIPIMAGE;
          if (cc.pco_debayer_algorithm)
            mode |= CONVERT_MODE_OUT_DOPCODEBAYER;
          },
        [&mode](auto& cc) {
          if (cc.sharpen)
            mode |= CONVERT_MODE_OUT_DOSHARPEN;
          if (cc.adaptive_sharpen)
            mode |= CONVERT_MODE_OUT_DOADSHARPEN;
          if (cc.flip_vertical)
            mode |= CONVERT_MODE_OUT_FLIPIMAGE;
          }
        }, conv_ctrl_map_.at(data_format));

      bool auto_minmax = std::visit([](auto& cc) {return cc.auto_minmax; }, conv_ctrl_map_.at(data_format));
      if (auto_minmax)
      {
        mode |= CONVERT_MODE_OUT_AUTOMINMAX;
        if (do_auto_minmax_map_.at(data_format))
        {
          setAutoMinMax(image_ref, data_format);
          do_auto_minmax_map_[data_format] = false;
        }
        if (checkMinMaxScale(data_format)) //check if we need to limit boundaries
          updateDisplaySettings(data_format);
      }

      if (data_format == DataFormat::BGRA8)
        mode |= CONVERT_MODE_OUT_RGB32;

      if (data_format == DataFormat::Mono8)
      {
        err = PCO_Convert16TO8(conv_map_[data_format], mode, calculateColorMode(roi),
          static_cast<int>(roi.width()), static_cast<int>(roi.height()),
          static_cast<WORD*>(image_ref.raw_data().first), static_cast<BYTE*>(image_ref.data().first));
      }
      else if (data_format == DataFormat::BGR16)
      {
        err = PCO_Convert16TOCOL16(conv_map_[data_format], mode, calculateColorMode(roi),
          static_cast<int>(roi.width()), static_cast<int>(roi.height()),
          static_cast<WORD*>(image_ref.raw_data().first), static_cast<WORD*>(image_ref.data().first));
      }
      else
      {
        if (isColored())
        {
          err = PCO_Convert16TOCOL(conv_map_[data_format], mode, calculateColorMode(roi),
            static_cast<int>(roi.width()), static_cast<int>(roi.height()),
            static_cast<WORD*>(image_ref.raw_data().first), static_cast<BYTE*>(image_ref.data().first));
        }
        else
        {
          err = PCO_Convert16TOPSEUDO(conv_map_[data_format], mode, calculateColorMode(roi),
            static_cast<int>(roi.width()), static_cast<int>(roi.height()),
            static_cast<WORD*>(image_ref.raw_data().first), static_cast<BYTE*>(image_ref.data().first));
        }
      }
      if (err)
        throw CameraException(err);
    }
  }

  //_________________________________________
  //get sdk, recorder, convert

  HANDLE Camera::sdk() const
  {
    return sdk_;
  }

  HANDLE Camera::rec() const
  {
    return rec_;
  }

  HANDLE Camera::conv(DataFormat data_format) const
  {
    if (conv_map_.find(data_format) == conv_map_.end())
      return nullptr;
    return conv_map_.at(data_format);
  }

  //_________________________________________
  // helper functions

  void Camera::close() noexcept
  {
    int err = PCO_NOERROR;
    for (auto& [format, handle] : conv_map_)
    {
      if (format == DataFormat::BGRA8)
        continue;
      err = PCO_ConvertDelete(handle);
      if (err)
        std::cout << "Close with error: " << CameraException(err).error_text() << "\n";
    }
    conv_map_.clear();

    if (rec_ != nullptr)
    {
      err = PCO_RecorderDelete(rec_);
      if (err)
        std::cout << "Close with error: " << CameraException(err).error_text() << "\n";
      rec_ = nullptr;
    }
    if (lens_ != nullptr)
    {
      err = PCO_CloseLensControl(sdk_);
      if (err)
        std::cout << "Close with error: " << CameraException(err).error_text() << "\n";
      lens_ = nullptr;
    }
    if (sdk_ != nullptr)
    {
      err = PCO_CloseCamera(sdk_);
      if (err)
        std::cout << "Close with error: " << CameraException(err).error_text() << "\n";
      sdk_ = nullptr;
    }
  }

  void Camera::scanForCamera(CameraInterface cam_interface)
  {
    int err = PCO_NOERROR;

    //create struct containing information about to-be-opened camera
    PCO_OpenStruct open_information{ sizeof(PCO_OpenStruct) };

    std::vector<CameraInterface> interfaces =
    { CameraInterface::FireWire     ,
      CameraInterface::CameraLinkMTX,
      CameraInterface::GenICam      ,
      CameraInterface::CameraLinkNAT,
      CameraInterface::GigE         ,
      CameraInterface::USB          ,
      CameraInterface::CameraLinkME4,
      CameraInterface::USB3         ,
      CameraInterface::CLHS         }; //in order of appearance

    //If an interface is specified, only scan for this one
    if (cam_interface != CameraInterface::Any)
    {
      interfaces.clear();
      interfaces.push_back(cam_interface);
    }

    for (auto& ci_iter : interfaces)
    {
      open_information.wInterfaceType = static_cast<WORD>(ci_iter);
      for (WORD i = 0; i < 64; i++)
      {
        open_information.wCameraNumAtInterface = i;
        open_information.wCameraNumber = i;
        err = PCO_OpenCameraEx(&sdk_, &open_information);
        if (err == PCO_NOERROR) //camera was found on port i
        {
          interface_ = ci_iter;
          return;
        }
        else if ((err & PCO_ERROR_DRIVER_NOTINIT) == PCO_ERROR_DRIVER_NOTINIT)  //no camera found on port i
        {
          continue;
        }
        else  //no camera found on this interface
        {
          break;
        }
      }
      //std::cout << "err: " << std::hex << err << std::dec << "\n";
    }
    throw CameraException(err);  //Throw if no camera found
  }

  bool Camera::hasRam()
  {
    if (camera_description_.dwGeneralCapsDESC1 & GENERALCAPS1_NO_RECORDER)
      return false;
    return true;
  }

  void Camera::switchToCamRam()
  {
    switchToCamRam(getCamRamSegment());
  }

  void Camera::switchToCamRam(WORD segment)
  {
    if (!hasRam())
      throw CameraException("Camera does not support CamRam!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_APPLICATION_FUNCTIONNOTSUPPORTED);

    int err = PCO_NOERROR;
    //______________________________________________________________
    //test if camera is currrently recording and if so then stop
    WORD state;
    err = PCO_GetRecordingState(sdk_, &state);
    if (err)
      throw CameraException(err);
    if (state == 1)
    {
      err = PCO_SetRecordingState(sdk_, 0);
      if (err)
        throw CameraException(err);
    }

    err = PCO_SetActiveRamSegment(sdk_, segment);
    if (err)
      throw CameraException(err);
    segment_ = segment;


    //______________________________________________________________
    //metadata mode can be activated for stored images, regardless of mode during recording.
    if ((camera_description_.dwGeneralCapsDESC1 & GENERALCAPS1_METADATA))
    {
      WORD metadata_size_dummy = 0;
      WORD metadata_version_dummy = 0;
      err = PCO_SetMetaDataMode(sdk_, METADATA_MODE_ON,
        &metadata_size_dummy, &metadata_version_dummy);
      if (err)
        throw CameraException(err);
      metadata_mode_on_ = true;
    }
    else
      metadata_mode_on_ = false;

    //______________________________________________________________
    //create recorder
    //___________________________________________
    //delete if already exists
    if (rec_ != nullptr)
    {
      err = PCO_RecorderStopRecord(rec_, sdk_);
      if (err)
        throw CameraException(err);
      err = PCO_RecorderDelete(rec_);
      if (err)
        throw CameraException(err);
      rec_ = nullptr;
    }

    //________________________________________________________
    // update all convert handles to sensor info that may have changed
    PCO_SensorInfo sensor_info = getSensorInfo();
    for (auto& [format, handle] : conv_map_)
    {
      err = PCO_ConvertSetSensorInfo(handle, &sensor_info);
      if (err)
        throw CameraException(err);
    }

    err = PCO_SetStorageMode(sdk_, STORAGE_MODE_RECORDER);
    if (err)
      throw CameraException(err);

    err = PCO_SetRecorderSubmode(sdk_, 1); //switch to ringbuffer mode
    if (err)
      throw CameraException(err);

    err = PCO_ArmCamera(sdk_);
    if ((err & PCO_ERROR_FIRMWARE_SEGMENT_TOO_SMALL) == PCO_ERROR_FIRMWARE_SEGMENT_TOO_SMALL)
    {
      // recorder handle not set.
      return;
    }
    else if (err)
    {
      throw CameraException(err);
    }

    WORD memory_mode = PCO_RECORDER_MODE_CAMRAM;
    recorder_buffer_size_ = 0;
    err = PCO_RecorderCreate(&rec_, &sdk_, nullptr, 1, memory_mode, nullptr, &recorder_buffer_size_);
    if (err)
      throw CameraException(err);


    DWORD max_images = recorder_buffer_size_;
    //___________________________________________
    //initialize recorder
    WORD memory_type = PCO_RECORDER_CAMRAM_SEQUENTIAL;
    if (recorder_buffer_size_)
    {
      err = PCO_RecorderInit(rec_, &max_images,
        1, //only one camera allowed
        memory_type, 0, nullptr, &segment_);
      if (err)
      {
        // ~Camera eventually deletes recorder
        throw CameraException(err);
      }
    }

    WORD ram_width = 0;
    WORD ram_height = 0;
    WORD meta_lines = 0;
    err = PCO_RecorderGetSettings(rec_, sdk_, NULL, &recorder_buffer_size_, &max_images, &ram_width, &ram_height, &meta_lines);
    if (err)
    {
      throw CameraException(err);
    }

    current_image_number_ = 0;

    hard_roi_.x0 = 1;
    hard_roi_.y0 = 1;
    hard_roi_.x1 = ram_width;
    hard_roi_.y1 = ram_height;
    double_image_factor_ = 1;


  }


  void Camera::setCamRamSegement(WORD segment)
  {
    if (!hasRam())
      throw CameraException("Camera does not support CamRam!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_APPLICATION_FUNCTIONNOTSUPPORTED);

    int err = PCO_SetActiveRamSegment(sdk_, segment);
    if (err)
      throw CameraException(err);

    return;
  }

  void Camera::calcTimeBase(double time, DWORD* time_calculated, WORD* base_calculated)
  {
    if (time < 0.004) //always take most fine granular, so rounding is not an issue
    {
      *base_calculated = 0;
      *time_calculated = static_cast<DWORD>(time * 1'000'000'000);
    }
    else if (time < 4)
    {
      *base_calculated = 1;
      *time_calculated = static_cast<DWORD>(time * 1'000'000);
    }
    else
    {
      *base_calculated = 2;
      *time_calculated = static_cast<DWORD>(time * 1'000);
    }
  }

  void Camera::setDelayExposureTime(double delay_time_s, double exposure_time_s)
  {
    DWORD time_exposure = UNDEF_DW;
    DWORD time_delay = UNDEF_DW;
    WORD base_exposure = UNDEF_W;
    WORD base_delay = UNDEF_W;

    if (std::isnan(exposure_time_s) || std::isnan(delay_time_s))
      throw CameraException("invalid exposure time or delay time: value is nan", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);

    if (exposure_time_s < 1e-9 * static_cast<double>(camera_description_.dwMinExposureDESC - camera_description_.dwMinExposureStepDESC))
      throw CameraException("invalid exposure time: smaller than min limit or negative", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);
    if (exposure_time_s > 1e-3 * static_cast<double>(camera_description_.dwMaxExposureDESC))
      throw CameraException("invalid exposure time: greater than max limit", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);
    if (delay_time_s < 1e-9 * static_cast<double>(camera_description_.dwMinDelayDESC))
      throw CameraException("invalid delay time: negative value", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);
    if (delay_time_s > 1e-3 * static_cast<double>(camera_description_.dwMaxDelayDESC))
      throw CameraException("invalid delay time: greater than max limit", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);

    calcTimeBase(exposure_time_s, &time_exposure, &base_exposure);
    calcTimeBase(delay_time_s, &time_delay, &base_delay);

    int err = PCO_NOERROR;
    err = PCO_SetDelayExposureTime(sdk_, time_delay, time_exposure, base_delay, base_exposure);
    if (err)
      throw CameraException(err);

    if (!isRecording())
    {
      //______________________________________________________________
      //load values into camera
      err = PCO_ArmCamera(sdk_);
      if (err)
        throw CameraException(err);
    }
  }

  int Camera::calculateColorMode(const Roi& roi)
  {
    if (!camera_description_.wColorPatternDESC)
      return 0;

    //Calculate overall ROI=hwOffset + swOffset + 1
    size_t roi_x0 = (hard_roi_.x0 - 1) + (roi.x0 - 1) + 1;
    size_t roi_y0 = (hard_roi_.y0 - 1) + (roi.y0 - 1) + 1;

    //Find position of red nibble
    //////////////////////////////////////////////////////////////////////////////
    int imodes[4][2][2] = { 3,2,1,0, // upper left is red
                            2,3,0,1, // upper left is green in red line
                            1,0,3,2, // upper left is green in blue line
                            0,1,2,3 };// upper left is blue

    int color_mode = camera_description_.wColorPatternDESC & 0x0F;
    if ((color_mode < 1) || (color_mode > 4))
      color_mode = 1;

    color_mode--;  // zero based index!
    color_mode = imodes[color_mode][roi_y0 & 0x01][roi_x0 & 0x01];

    return color_mode;
  }

  PCO_Recorder_CompressionParams Camera::getCompressionParams()
  {
    PCO_Recorder_CompressionParams params{};
    //currently use default values
    params.dGainK = (1 / 0.46);
    params.dDarkNoise_e = 1.5;
    params.dDSNU_e = 0.3;
    params.dPRNU_pct = 0.34;
    params.dLightSourceNoise_pct = 0.0;

    return params;
  }

  void Camera::setAutoMinMax(Image& image_ref, DataFormat data_format)
  {
    int err = PCO_NOERROR;
    PCO_Display display{ sizeof(PCO_Display) };
    err = PCO_ConvertGetDisplay(conv_map_[data_format], &display);
    if (err)
      throw CameraException(err);
    auto max = std::max_element(image_ref.raw_vector_16bit().begin(), image_ref.raw_vector_16bit().end());
    auto min = std::min_element(image_ref.raw_vector_16bit().begin(), image_ref.raw_vector_16bit().end());
    if (*min == *max && *min != 0)
      --(*min);
    display.iScale_max = static_cast<int>(*max);
    display.iScale_min = static_cast<int>(*min);
    err = PCO_ConvertSetDisplay(conv_map_[data_format], &display);
    if (err)
      throw CameraException(err);
  }

  bool Camera::checkMinMaxScale(DataFormat data_format)
  {
    if (!isColored())
      return false;  //No need to check anything
    if ((data_format != DataFormat::BGR8) && (data_format != DataFormat::BGRA8))
      return false;  //No need to check anything

    ConvertControlColor& cc_color = std::get<ConvertControlColor>(conv_ctrl_map_.at(data_format));
    float red_max, green_max, blue_max;
    int err = PCO_GetMaxLimit(&red_max, &green_max, &blue_max, (float)cc_color.color_temperature, (float)cc_color.color_tint, camera_description_.wDynResDESC);
    if (err)
      throw CameraException(err);

    float colMax = std::max({ red_max, green_max, blue_max });
    float colMin = std::min({ red_max, green_max, blue_max });

    conv_maxmax_ = (int)((float)((1 << camera_description_.wDynResDESC) - 1) / colMax);
    conv_maxmin_ = (int)((float)cc_color.max_limit * colMin);
    conv_minmax_ = conv_maxmin_;
    conv_maxmin_ = (cc_color.max_limit - conv_maxmin_) + cc_color.min_limit;
    if (conv_minmax_ < cc_color.min_limit)
      conv_minmax_ = cc_color.min_limit + 1;

    bool needsUpdate = false;
    if (cc_color.max_limit > conv_maxmax_)
    {
      cc_color.max_limit = conv_maxmax_;
      needsUpdate = true;
    }
    if (cc_color.max_limit < conv_maxmin_)
    {
      cc_color.max_limit = conv_maxmin_;
      needsUpdate = true;
    }
    if (cc_color.min_limit > conv_minmax_)
    {
      cc_color.min_limit = conv_minmax_;
      needsUpdate = true;
    }

    return needsUpdate;
  }

  void Camera::updateDisplaySettings(DataFormat data_format)
  {
    if (conv_map_.find(data_format) == conv_map_.end())
    {
      throw CameraException("Data format is not supported by the camera!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);
    }

    PCO_Display display{ sizeof(PCO_Display) };
    DWORD err = PCO_ConvertGetDisplay(conv_map_.at(data_format), &display);
    if (err)
      throw CameraException(err);

    display.iContrast = 0;

    std::visit(overloaded{
      [&display](ConvertControlMono& cc_mono) {
          display.iGamma = static_cast<int>(cc_mono.gamma * 100.0);
          display.iContrast = cc_mono.contrast;
          if (!cc_mono.auto_minmax)
          {
            display.iScale_min = cc_mono.min_limit;
            display.iScale_max = cc_mono.max_limit;
          }
        },
      [&display](ConvertControlColor& cc_color) {

          display.iGamma = static_cast<int>(cc_color.gamma * 100.0);
          display.iContrast = cc_color.contrast;
          display.iColor_temp = cc_color.color_temperature;
          display.iColor_saturation = cc_color.color_saturation;
          display.iColor_vibrance = cc_color.color_vibrance;
          display.iColor_tint = cc_color.color_tint;
          if (!cc_color.auto_minmax)
          {
            display.iScale_min = cc_color.min_limit;
            display.iScale_max = cc_color.max_limit;
          }
        },
      [&display](ConvertControlPseudoColor& cc_color) {
          display.iGamma = static_cast<int>(cc_color.gamma * 100.0);
          display.iContrast = cc_color.contrast;
          display.iColor_temp = cc_color.color_temperature;
          display.iColor_saturation = cc_color.color_saturation;
          display.iColor_vibrance = cc_color.color_vibrance;
          display.iColor_tint = cc_color.color_tint;
          if (!cc_color.auto_minmax)
          {
            display.iScale_min = cc_color.min_limit;
            display.iScale_max = cc_color.max_limit;
          }
        }
      }, conv_ctrl_map_.at(data_format));

    err = PCO_ConvertSetDisplay(conv_map_.at(data_format), &display);
    if (err)
      throw CameraException(err);
  }

  bool Camera::recIsCamRam()
  {
    DWORD mode_and_type = 0;
    int err = PCO_RecorderGetSettings(rec_, sdk_, &mode_and_type, nullptr, nullptr, nullptr, nullptr, nullptr);
    if (err)
      throw CameraException(err);

    if ((HIWORD(mode_and_type) & PCO_RECORDER_MODE_CAMRAM) == PCO_RECORDER_MODE_CAMRAM)
      return true;
    return false;
  }

  WORD Camera::getCamRamSegment()
  {
    if (!hasRam())
      throw CameraException("Camera does not support CamRam!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_APPLICATION_FUNCTIONNOTSUPPORTED);

    WORD seg = 0;
    int err = PCO_GetActiveRamSegment(sdk_, &seg);
    if (err)
      throw CameraException(err);

    return seg;
  }

  DWORD Camera::getCamRamMaxImages()
  {
    if (!hasRam())
      throw CameraException("Camera does not support CamRam!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_APPLICATION_FUNCTIONNOTSUPPORTED);

    int err = PCO_NOERROR;
    if (!rec_)
    {
      DWORD valid_images = 0;
      DWORD max_images = 0;
      err = PCO_GetNumberOfImagesInSegment(sdk_, segment_, &valid_images, &max_images);
      if (err)
        throw CameraException(err);
      return max_images; // always 0, since otherwise recorder exists.
    }

    if (recIsCamRam())
    {
      return recorder_buffer_size_;
    }

    DWORD valid_images = 0;
    DWORD max_images = 0;
    err = PCO_GetNumberOfImagesInSegment(sdk_, segment_, &valid_images, &max_images);
    if (err)
      throw CameraException(err);
    return max_images;

  }
  DWORD Camera::getCamRamNumImages()
  {
    if (!hasRam())
      throw CameraException("Camera does not support CamRam!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_APPLICATION_FUNCTIONNOTSUPPORTED);

    int err = PCO_NOERROR;
    if (!rec_)
    {
      return 0;
    }

    if (recIsCamRam())
    {
      DWORD valid_images = 0;
      DWORD max_images = 0;
      err = PCO_GetNumberOfImagesInSegment(sdk_, segment_, &valid_images, &max_images);
      if (err)
        throw CameraException(err);
      return valid_images; // can be queried while other record mode is running.
    }

    DWORD image_count = UNDEF_DW;
    err = PCO_RecorderGetStatus(rec_, sdk_, nullptr, nullptr, nullptr,
      &image_count,     //get index of last recorded image
      nullptr, nullptr, nullptr, nullptr, nullptr);
    if (err)
      throw CameraException(err);

    return image_count;
  }

  void Camera::setCamRamAllocation(std::vector<DWORD> percents)
  {

    if (!hasRam())
      throw CameraException("Camera does not support CamRam!", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);

    PCO_Storage storage{ sizeof(PCO_Storage) };
    int err = PCO_GetStorageStruct(sdk_, &storage);
    if (err)
      throw CameraException(err);

    percents.resize(4);

    if (std::accumulate(percents.begin(), percents.end(), 0UL) > 100)
      throw CameraException("Sum of percentages is greater than 100%", PCO_ERROR_HIGHLEVELSDK | PCO_ERROR_WRONGVALUE);

    std::for_each(percents.begin(), percents.end(), [&](DWORD& ram_in_percent) {
      ram_in_percent = ram_in_percent * (storage.dwRamSize / 100); // consider potential DWORD overflow
      });

    err = PCO_SetCameraRamSegmentSize(sdk_, percents.data());
    if (err)
      throw CameraException(err);
  }



}
