#pragma once
#include "defs.h"
#include "cameraexception.h"
#include "image.h"

namespace pco
{
  class Camera
  {
  public:

    //_________________________________________
    // constructor + destructor

    Camera(CameraInterface cam_interface = CameraInterface::Any);
    ~Camera();

    //_________________________________________
    // camera properties

    std::string getName() const;
    RawFormat getRawFormat() const;
    DWORD getSerial() const;
    CameraInterface getInterface() const;
    bool isRecording() noexcept;
    bool isColored() const;

    //_________________________________________
    // control camera

    Description getDescription() const;

    void defaultConfiguration();
    Configuration getConfiguration() const;
    void setConfiguration(Configuration config);

    double getExposureTime();
    void setExposureTime(double exposure_time_s);

    double getDelayTime();
    void setDelayTime(double delay_time_s);

    //_________________________________________
    // handle recorder

    void record(DWORD num_images = 1, RecordMode record_mode = RecordMode::sequence, std::filesystem::path file_path = "");
    void stop();

    void waitForFirstImage(bool delay = true, double timeout_s = NAN);
    void waitForNewImage(bool delay = true, double timeout_s = NAN);
    size_t getRecordedImageCount() const;
    size_t getRecorderBufferSize() const;

    //_________________________________________
    // convert

    ConvertControl getConvertControl(DataFormat data_format);
    void setConvertControl(DataFormat data_format, ConvertControl variant_settings);

    // special convert functions:

    void loadLut(DataFormat data_format, std::filesystem::path lut_file);
    void adaptWhiteBalance(Image& image);
    void adaptWhiteBalance(Image& image, Roi roi);

    //_________________________________________
    // images

    void image(Image& image_ref, DWORD image_index = 0, DataFormat data_format = DataFormat::Mono16, PCO_Recorder_CompressionParams* comp_params = nullptr);
    void image(Image& image_ref, Roi roi, DWORD image_index = 0, DataFormat data_format = DataFormat::Mono16, PCO_Recorder_CompressionParams* comp_params = nullptr);

    void images(std::vector<Image>& images, DataFormat data_format = DataFormat::Mono16, size_t start_index = 0, PCO_Recorder_CompressionParams* comp_params = nullptr);
    void images(std::vector<Image>& images, Roi roi, DataFormat data_format = DataFormat::Mono16, size_t start_index = 0, PCO_Recorder_CompressionParams* comp_params = nullptr);

    void imageAverage(Image& image_ref, DataFormat data_format = DataFormat::Mono16);
    void imageAverage(Image& image_ref, Roi roi, DataFormat = DataFormat::Mono16);

    //_________________________________________
    // camram control for memory cameras

    bool hasRam();
    void switchToCamRam(); // take active segment; switch readout for image()
    void switchToCamRam(WORD segment); // switch readout for image()
    WORD getCamRamSegment();
    DWORD getCamRamMaxImages();
    DWORD getCamRamNumImages();
    void setCamRamAllocation(std::vector<DWORD> percents);


    //_________________________________________
    // lib handels for sdk, recorder, convert

    HANDLE sdk() const;
    HANDLE rec() const;
    HANDLE conv(DataFormat data_format) const;

  private:

    //handles

    HANDLE sdk_ = nullptr;
    HANDLE rec_ = nullptr;
    HANDLE lens_ = nullptr; //lens control

    //camera infos

    PCO_Description camera_description_;
    PCO_CameraType camera_type_;
    std::string camera_name_;
    CameraInterface interface_ = CameraInterface::Any;

    // volatile camera setup

    Roi hard_roi_;  //ROI as set within camera != ROI of image; hard_roi_ can be seen as sensor size
    WORD double_image_factor_ = 1;

    DWORD current_image_number_ = 0;
    DWORD recorder_buffer_size_ = UNDEF_DW; // camram: max images in segment

    bool metadata_mode_on_ = false;
    bool timestamp_mode_binary_ = false; // If ts_mode is only ascii, value is stored in image, camera will throw CameraException(err) if asked for it anyway

    // image convert

    std::unordered_map<DataFormat, HANDLE> conv_map_;
    std::unordered_map<DataFormat, ConvertControl> conv_ctrl_map_;
    std::unordered_map<DataFormat, bool> do_auto_minmax_map_;

    int conv_maxmax_ = 0;
    int conv_maxmin_ = 0;
    int conv_minmax_ = 0;

    //_________________________________________
    // helper functions

    void close() noexcept;
    void scanForCamera(CameraInterface cam_interface);

    WORD segment_ = 0; // = not set
    void setCamRamSegement(WORD segment);

    static void calcTimeBase(double time, DWORD* time_calculated, WORD* base_calculated);
    void setDelayExposureTime(double delay_time_s, double exposure_time_s);

    int calculateColorMode(const Roi& roi);
    PCO_SensorInfo getSensorInfo() const;
    PCO_Recorder_CompressionParams getCompressionParams();
    void setAutoMinMax(Image& image_ref, DataFormat data_format);
    bool checkMinMaxScale(DataFormat data_format);
    void updateDisplaySettings(DataFormat data_format);

    bool recIsCamRam();

  };

};
